@@ -28,6 +28,7 @@ import unittest
 from Crypto.Hash import SHA
 import paramiko.util
 
+from util import ParamikoTest
 
 test_config_file = """\
 Host *
@@ -58,17 +59,7 @@ BGQ3GQ/Fc7SX6gkpXkwcZryoi4kNFhHu5LvHcZPdxXV1D+uTMfGS1eyd2Yz/DoNWXNAl8TI0cAsW\
 from paramiko import *
 
 
-class UtilTest (unittest.TestCase):
-
-    assertTrue = unittest.TestCase.failUnless   # for Python 2.3 and below
-    assertFalse = unittest.TestCase.failIf      # for Python 2.3 and below
-
-    def setUp(self):
-        pass
-
-    def tearDown(self):
-        pass
-    
+class UtilTest(ParamikoTest):
     def test_1_import(self):
         """
         verify that all the classes can be imported from paramiko.
